module psf

  use param
  use dal
  use cal
  use emldata
  
  implicit none

  private

  public :: psf1, psf2, psf3, psf4, psf5, psf6

contains

  function psfoff (ene, eps, x, sig, ierr) result (result_psfoff) 

!  calculates XMM off-axis PSF (normalized surface brightness)
!
!  author : HEB        date: 15-NOV-1996 09:21 
!  modifi : GRH        date: 15-NOV-1996 20:07  off-axis from ABRIXAS 
!  modifi : GRH        date: 19-NOV-1996 09:50  XMM simulations included
!
!  The XMM QM PSF has been fit by Mike Watson with 4 Gaussians  
!  GRH has normalized the PSF components and added a mirror-blur term in
!  quadrature as a function of off axis angle, derived from XMM 
!  simulations performed by P.Friedrich using the ABRIXAS sim code.
!
!   call_var.          type I/O description
!  ENE                 R   I   Energy [keV]
!  EPS                  R   I      off-axis angle [arc min]
!  X                   R   I   angle from target position [arcsec]
!
!  IERR                I   O   = 0 no error 
!                              = 1 energy outside bounds 0.07-3 keV
!                              = 3 
!
!     OUTPUT:
!             PSFOFF      surface brightness of point spread function,
!                         normalized, so that  Integral 2*PI*r*dr*f
!                         from 0 to infinity = 1 [1/arcsec2]

    real          result_psfoff
    real          ene, eps, x, sig
    integer       ierr
    real          norm(4),sigma(4),sum,offsg,pi,sig2,rnorm,sigtel
    integer       i

! MGW data norm  /0.75337,0.14463,0.79143e-2,0.75355e-3/,
! following are the norms after integrating each component to 1.00
    data norm  /0.4169,0.3470,0.1279,0.1082/, &
         sigma /4.1450,8.6218,22.372,66.698/, &
         offsg /0.0324/, & 
         pi    /3.141592654/

    sum=0.

! off-axis component of telescope    
    sigtel=offsg*min(eps,18.0)**2
    do i=1,4
      sig2=0.5/(sigma(i)**2+sigtel**2+sig**2)
      rnorm=norm(i)*sig2/pi
      sum=sum+rnorm*exp(-sig2*(x)**2)
    end do

    result_psfoff = sum

  end function psfoff

  real function psf1 (energy, iband, x1, y1, x0, y0, xcen, ycen, sig, ecut) &
       result (result_psf1)

    real, intent(in) :: x0, y0, x1, y1, xcen, ycen, sig, ecut
    real, dimension(max_ima), intent(in) :: energy
    integer :: iband

    real, parameter :: pi = 3.14159265, sec2mm = 7500.0*pi/(180.0*3600.0)
    real(kind=double) :: E, theta, phi, E_old, theta_old, phi_old, val
    logical :: first = .true.
    integer :: i,ix,iy,ixx,iyy,lx,ly,kx,ky,kk,np,np2,isig3,isig3sq
    integer, dimension(max_ima) :: nx, ny
    type (PsfT) :: ppsf
    type(PsfBinSizeT), dimension(max_ima) :: bsiz
    type(PSFwindowT) :: window
    type(PsfValidityRangesT) :: vrange 
    real(kind=single), dimension(:,:), pointer :: pima
    real, dimension(:,:,:), allocatable :: sima 
    real, dimension(0:100,0:100) :: gengen
    real :: x,y,dx,dy,ddx,ddy,xd0,yd0,xx,yy, &
            sum,winr,sig0,sig2,rnorm,pfac
    real :: sigold = -1.0
    integer :: sxlo,sxhi,sylo,syhi

    save

    xd0=x0-xcen
    yd0=y0-ycen
    theta=sqrt(xd0**2+yd0**2)*pixim
    phi=atan2(xd0,yd0) + pi/2.0

    if (first) then
      call CAL_setState1(instrument=EPN, &
                         AccuracyLevel=ACCURACY_MEDIUM)
      allocate (sima(nim,250,250))
      gengen(0,0) = 1.0
    end if 

    vrange = CAL_PsfValidityRanges()
    theta = max(min(theta,vrange%thetaMax-1.0),vrange%thetaMin+1.0)

    if (theta<theta_old-30d0.or.theta>theta_old+30d0 .or. &
        phi<phi_old-0.1.or.phi>phi_old+0.1 .or. first) then 

      np = nint((pixim-1.0)/2.0)
      np2= (2*np+1)**2 

      winr = int(ecut)+2.0
      winr=  min(winr*pixim,100.0)*sec2mm

      window%halfX = winr
      window%halfY = winr

      do i=1,nim
        E = energy(i)
        E = max(min(E,vrange%eMax-1.0),vrange%eMin+1.0)
        ppsf = Cal_getPSF(E, theta, phi) 
        bsiz(i) = CAL_psfBINSIZE(ppsf) 
        call CAL_psfImage (ppsf, window, pima) 
        nx(i) = size(pima,1)
        ny(i) = size(pima,1)
        do iy=np,ny(i)-1-np 
          do ix=np,nx(i)-1-np
            sum=0.0
            do ixx=ix-np,ix+np
              do iyy=iy-np,iy+np
                sum=sum+pima(ixx,iyy)
              end do
            end do
            sima(i,ix,iy)=sum/np2
          end do
        end do 
        call CAL_releaseMemory(pima)
      end do 

      first = .false. 
      theta_old = theta
      phi_old = phi
      E_old = E 

    end if

    xx = ((x1-x0)*pixim*sec2mm+window%halfX)/bsiz(iband)%binX-0.5
    yy = ((y1-y0)*pixim*sec2mm+window%halfY)/bsiz(iband)%binY-0.5   

    pfac=pixim*sec2mm/bsiz(iband)%binX
    isig3 = nint(min(3.*sig,ecut))
    isig3sq = isig3**2

    if (sig/=sigold) then
      do ly = 0,isig3
        do lx = 1,ly
          if (lx**2+ly**2<=isig3sq) then        
            x=lx
            y=ly
            if (lx/=ly) then
              gengen(lx,ly) = gen(0.0,x-0.5,x+0.5,sig)*gen(0.0,y-0.5,y+0.5,sig)
            else
              gengen(lx,lx) = gen(0.0,x-0.5,x+0.5,sig)**2
            end if
          end if 
        end do
      end do
      sigold=sig
    end if
      
    sxlo = max(-isig3,int((1-xx)/pfac))
    sxhi = min(isig3,int((nx(iband)-2-xx)/pfac)+1)
    sylo = max(-isig3,int((1-yy)/pfac))
    syhi = min(isig3,int((ny(iband)-2-yy)/pfac)+1)

    sum = 0.
    do ly = sylo,syhi   ! -isig3,isig3
      do lx = sxlo,sxhi ! -isig3,isig3
        ! if (lx**2+ly**2<=isig3sq) then
          x=xx+lx*pfac
          y=yy+ly*pfac 
          ix = min(max(int(x),1),nx(iband)-2)
          iy = min(max(int(y),1),ny(iband)-2)
          dx=x-ix
          dy=y-iy
          ddx=1.-dx
          ddy=1.-dy        

          val = ddx*ddy*sima(iband,ix  ,iy)   + &
                ddx* dy*sima(iband,ix  ,iy+1) + &
                 dx*ddy*sima(iband,ix+1,iy)   + &
                 dx* dy*sima(iband,ix+1,iy+1)
        
          kx = abs(lx)
          ky = abs(ly)
          if (kx>ky) then
            kk=kx
            kx=ky
            ky=kk
          end if 
          sum = sum + val * gengen (kx,ky)

        ! end if
      end do
    end do 
    
    result_psf1 = sum * (sec2mm*pixim)**2

  end function psf1

  real function psf2 (energy, iband, x1, y1, x0, y0, xcen, ycen, sig, ecut) &
       result (result_psf2)

    real, intent(in) :: x0, y0, x1, y1, xcen, ycen, sig, ecut
    real, dimension(max_ima), intent(in) :: energy
    integer :: iband

    real, parameter :: pi = 3.14159265, sec2mm = 7500.0*pi/(180.0*3600.0)
    real(kind=double) :: E, theta, phi, E_old, theta_old, phi_old, val
    logical :: first = .true.
    integer :: i,ix,iy,ixx,iyy,lx,ly,kx,ky,kk,np,np2,isig3,isig3sq
    integer, dimension(max_ima) :: nx, ny
    type (PsfT) :: ppsf
    type(PsfBinSizeT), dimension(max_ima) :: bsiz
    type(PSFwindowT) :: window
    type(PsfValidityRangesT) :: vrange 
    real(kind=single), dimension(:,:), pointer :: pima
    real, dimension(:,:,:), allocatable :: sima 
    real, dimension(0:100,0:100) :: gengen
    real :: x,y,dx,dy,ddx,ddy,xd0,yd0,xx,yy, &
            sum,winr,sig0,sig2,rnorm,pfac
    real :: sigold = -1.0
    integer :: sxlo,sxhi,sylo,syhi

    save

    xd0=x0-xcen
    yd0=y0-ycen
    theta=sqrt(xd0**2+yd0**2)*pixim
    phi=atan2(xd0,yd0) + pi/2.0

    if (first) then
      call CAL_setState1(instrument=EMOS1, &
                         AccuracyLevel=ACCURACY_MEDIUM)
      allocate (sima(nim,250,250))
      gengen(0,0) = 1.0
    end if 

    vrange = CAL_PsfValidityRanges()
    theta = max(min(theta,vrange%thetaMax-1.0),vrange%thetaMin+1.0)

    if (theta<theta_old-30d0.or.theta>theta_old+30d0 .or. &
        phi<phi_old-0.1.or.phi>phi_old+0.1 .or. first) then 

      np = nint((pixim-1.0)/2.0)
      np2= (2*np+1)**2 

      winr = int(ecut)+2.0
      !print *,"winr1", winr
      winr=  min(winr*pixim,100.0)*sec2mm
      !print *,"winr2",winr 

      window%halfX = winr
      window%halfY = winr

      do i=1,nim
        E = energy(i)
        E = max(min(E,vrange%eMax-1.0),vrange%eMin+1.0)
        ppsf = Cal_getPSF(E, theta, phi) 
        bsiz(i) = CAL_psfBINSIZE(ppsf) 
        call CAL_psfImage (ppsf, window, pima) 
        nx(i) = size(pima,1)
        ny(i) = size(pima,1)
        !print *,"nxy",nx(i),ny(i)
        do iy=np,ny(i)-1-np 
          do ix=np,nx(i)-1-np
            sum=0.0
            do ixx=ix-np,ix+np
              do iyy=iy-np,iy+np
                sum=sum+pima(ixx,iyy)
              end do
            end do
            sima(i,ix,iy)=sum/np2
          end do
        end do 
        call CAL_releaseMemory(pima)
      end do 

      first = .false. 
      theta_old = theta
      phi_old = phi
      E_old = E 

    end if

    xx = ((x1-x0)*pixim*sec2mm+window%halfX)/bsiz(iband)%binX-0.5
    yy = ((y1-y0)*pixim*sec2mm+window%halfY)/bsiz(iband)%binY-0.5   

    pfac=pixim*sec2mm/bsiz(iband)%binX
    isig3 = nint(min(3.*sig,ecut))
    isig3sq = isig3**2

    if (sig/=sigold) then
      do ly = 0,isig3
        do lx = 1,ly
          if (lx**2+ly**2<=isig3sq) then        
            x=lx
            y=ly
            if (lx/=ly) then
              gengen(lx,ly) = gen(0.0,x-0.5,x+0.5,sig)*gen(0.0,y-0.5,y+0.5,sig)
            else
              gengen(lx,lx) = gen(0.0,x-0.5,x+0.5,sig)**2
            end if
          end if 
        end do
      end do
      sigold=sig
    end if
      
    sxlo = max(-isig3,int((1-xx)/pfac))
    sxhi = min(isig3,int((nx(iband)-2-xx)/pfac)+1)
    sylo = max(-isig3,int((1-yy)/pfac))
    syhi = min(isig3,int((ny(iband)-2-yy)/pfac)+1)

    sum = 0.
    do ly = sylo,syhi    ! -isig3,isig3
      do lx = sxlo,sxhi  ! -isig3,isig3
        ! if (lx**2+ly**2<=isig3sq) then
          x=xx+lx*pfac
          y=yy+ly*pfac 
          ix = min(max(int(x),1),nx(iband)-2)
          iy = min(max(int(y),1),ny(iband)-2)
          dx=x-ix
          dy=y-iy
          ddx=1.-dx
          ddy=1.-dy        

          val = ddx*ddy*sima(iband,ix  ,iy)   + &
                ddx* dy*sima(iband,ix  ,iy+1) + &
                 dx*ddy*sima(iband,ix+1,iy)   + &
                 dx* dy*sima(iband,ix+1,iy+1)
        
          kx = abs(lx)
          ky = abs(ly)
          if (kx>ky) then
            kk=kx
            kx=ky
            ky=kk
          end if 
          sum = sum + val * gengen (kx,ky)

        ! end if
      end do
    end do 
    
    result_psf2 = sum * (sec2mm*pixim)**2

  end function psf2

  real function psf3 (energy, iband, x1, y1, x0, y0, xcen, ycen, sig, ecut) &
       result (result_psf3)

    real, intent(in) :: x0, y0, x1, y1, xcen, ycen, sig, ecut
    real, dimension(max_ima), intent(in) :: energy
    integer :: iband

    real, parameter :: pi = 3.14159265, sec2mm = 7500.0*pi/(180.0*3600.0)
    real(kind=double) :: E, theta, phi, E_old, theta_old, phi_old, val
    logical :: first = .true.
    integer :: i,ix,iy,ixx,iyy,lx,ly,kx,ky,kk,np,np2,isig3,isig3sq
    integer, dimension(max_ima) :: nx, ny
    type (PsfT) :: ppsf
    type(PsfBinSizeT), dimension(max_ima) :: bsiz
    type(PSFwindowT) :: window
    type(PsfValidityRangesT) :: vrange 
    real(kind=single), dimension(:,:), pointer :: pima
    real, dimension(:,:,:), allocatable :: sima 
    real, dimension(0:100,0:100) :: gengen
    real :: x,y,dx,dy,ddx,ddy,xd0,yd0,xx,yy, &
            sum,winr,sig0,sig2,rnorm,pfac
    real :: sigold = -1.0
    integer :: sxlo,sxhi,sylo,syhi

    save

    xd0=x0-xcen
    yd0=y0-ycen
    theta=sqrt(xd0**2+yd0**2)*pixim
    phi=atan2(xd0,yd0) + pi/2.0

    if (first) then
      call CAL_setState1(instrument=EMOS2, &
                         AccuracyLevel=ACCURACY_MEDIUM)
      allocate (sima(nim,250,250))
      gengen(0,0) = 1.0
    end if 

    vrange = CAL_PsfValidityRanges()
    theta = max(min(theta,vrange%thetaMax-1.0),vrange%thetaMin+1.0)

    if (theta<theta_old-30d0.or.theta>theta_old+30d0 .or. &
        phi<phi_old-0.1.or.phi>phi_old+0.1 .or. first) then 

      np = nint((pixim-1.0)/2.0)
      np2= (2*np+1)**2 

      winr = int(ecut)+2.0
      winr=  min(winr*pixim,100.0)*sec2mm

      window%halfX = winr
      window%halfY = winr

      do i=1,nim
        E = energy(i)
        E = max(min(E,vrange%eMax-1.0),vrange%eMin+1.0)
        ppsf = Cal_getPSF(E, theta, phi) 
        bsiz(i) = CAL_psfBINSIZE(ppsf) 
        call CAL_psfImage (ppsf, window, pima) 
        nx(i) = size(pima,1)
        ny(i) = size(pima,1)
        do iy=np,ny(i)-1-np 
          do ix=np,nx(i)-1-np
            sum=0.0
            do ixx=ix-np,ix+np
              do iyy=iy-np,iy+np
                sum=sum+pima(ixx,iyy)
              end do
            end do
            sima(i,ix,iy)=sum/np2
          end do
        end do 
        call CAL_releaseMemory(pima)
      end do 

      first = .false. 
      theta_old = theta
      phi_old = phi
      E_old = E 

    end if

    xx = ((x1-x0)*pixim*sec2mm+window%halfX)/bsiz(iband)%binX-0.5
    yy = ((y1-y0)*pixim*sec2mm+window%halfY)/bsiz(iband)%binY-0.5   

    pfac=pixim*sec2mm/bsiz(iband)%binX
    isig3 = nint(min(3.*sig,ecut))
    isig3sq = isig3**2

    if (sig/=sigold) then
      do ly = 0,isig3
        do lx = 1,ly
          if (lx**2+ly**2<=isig3sq) then        
            x=lx
            y=ly
            if (lx/=ly) then
              gengen(lx,ly) = gen(0.0,x-0.5,x+0.5,sig)*gen(0.0,y-0.5,y+0.5,sig)
            else
              gengen(lx,lx) = gen(0.0,x-0.5,x+0.5,sig)**2
            end if
          end if 
        end do
      end do
      sigold=sig
    end if
      
    sum = 0.

    sxlo = max(-isig3,int((1-xx)/pfac))
    sxhi = min(isig3,int((nx(iband)-2-xx)/pfac)+1)
    sylo = max(-isig3,int((1-yy)/pfac))
    syhi = min(isig3,int((ny(iband)-2-yy)/pfac)+1)

    do ly = sylo,syhi   ! -isig3,isig3
      do lx = sxlo,sxhi ! -isig3,isig3
        ! if (lx**2+ly**2<=isig3sq) then
          x=xx+lx*pfac
          y=yy+ly*pfac 
          ix = min(max(int(x),1),nx(iband)-2)
          iy = min(max(int(y),1),ny(iband)-2)
          dx=x-ix
          dy=y-iy
          ddx=1.-dx
          ddy=1.-dy        

          val = ddx*ddy*sima(iband,ix  ,iy)   + &
                ddx* dy*sima(iband,ix  ,iy+1) + &
                 dx*ddy*sima(iband,ix+1,iy)   + &
                 dx* dy*sima(iband,ix+1,iy+1)
        
          kx = abs(lx)
          ky = abs(ly)
          if (kx>ky) then
            kk=kx
            kx=ky
            ky=kk
          end if 
          sum = sum + val * gengen (kx,ky)

        ! end if
      end do
    end do 
    
    result_psf3 = sum * (sec2mm*pixim)**2

  end function psf3

  real function psf4 (energy, iband, x1, y1, x0, y0, xcen, ycen, sig, ecut) &
       result (result_psf4)

    real, intent(in) :: x0, y0, x1, y1, xcen, ycen, sig, ecut
    real, dimension(max_ima), intent(in) :: energy
    integer :: iband, ierr

    real    :: x, eps, xsig, xd0, yd0   
     
    xd0=x0-xcen
    yd0=y0-ycen
    x = sqrt ((x1-x0)**2+(y1-y0)**2) * pixim
    xsig = sig * pixim
    eps = sqrt(xd0**2+yd0**2)*pixim/60.
    result_psf4 = psfoff (1.0, eps, x, xsig, ierr) * pixim**2

  end function psf4

  real function psf5 (ene, offax, posang, x1, y1, x0, y0, sig)

    real, intent(in) :: ene, offax, posang, x0, y0, x1, y1, sig
    psf5 = 0.

  end function psf5

  real function psf6 (ene,offax,posang,x1,y1,x0,y0,sig)

    real, intent(in) :: ene, offax, posang, x0, y0, x1, y1, sig
    psf6 = 0.

  end function psf6

  real function gen(x0,x1,x2,sig)

    real x0,x1,x2,sig,fsig
                                                                       
    gen=0.                                                           
    if (sig.le.0.00001) sig=0.00001
    fsig = 1.0/(1.414214*sig)                                        
    gen = (errf((x2-x0)*fsig)-errf((x1-x0)*fsig))/2.0

  end function gen     

  real function errf (xx)

    real t,x,xx

    x=abs(xx)                                                        
    errf=1.0
                                                         
    if (x<=5.0) then                                            
       t=1.0/(1.0+0.47047*x)                                            
       errf=1.0-(0.3480242-(0.0958798-0.7478556*t)*t)*t*exp(-(x*x))
     end if   
                                                                    
     if (xx<0.0) errf=-errf                                        

   end function errf     

end module psf





